import json
import pandas as pd

def get_exposure_times(log_file, t0):
    data = []
    seen_layers = set()  # to track already recorded layers
    # Read and extract temperature/time
    with open(log_file, 'r') as f:
        for line in f:
            # Skip any line that contains "Retry"
            if "Retry" in line:
                continue
            if 'freemelt/0/BuildProcess/0/BuildStatus/Trace/Activity' in line:
                try:
                    json_part = line.split(' ', 1)[1]
                    values = json.loads(json_part)
                    temperature = values.get('current_activity')
                    if temperature and temperature.startswith("Exposing OBP files"):
                        parts = temperature.split("layer")
                        if len(parts) > 1:
                            layer_number = parts[1].strip(" .")
                            # Skip if layer was already seen
                            if layer_number in seen_layers:
                                continue
                            seen_layers.add(layer_number)
                            time_ns = int(values.get('time'))  # Make sure it's an int
                            time = time_ns / 1e9
                            time -= t0  # Convert to seconds
                            if time > 0:
                                data.append([time, layer_number])
                except (IndexError, json.JSONDecodeError, TypeError, ValueError):
                    continue  # skip malformed lines
    df = pd.DataFrame(data, columns=['Time (s)', 'Exposure Layer'])
    # Optional: sort by layer number
    df['Exposure Layer'] = df['Exposure Layer'].astype(int)
    df = df.sort_values('Exposure Layer').reset_index(drop=True)
    return df

def get_power_times(log_file, t0):
    data = []
    with open(log_file, 'r') as f:
        for line in f:
            if 'freemelt/0/ElectronService/0/HighVoltageGenerator/Type/Current' in line:
                #print(line)
                try:
                    json_part = line.split(' ', 1)[1]
                    values = json.loads(json_part)
                    temperature = values.get('CurrentDemand')
                    time_ns = int(values.get('time'))  # Make sure it's an int
                    if time_ns > t0:
                        current_time = (time_ns - t0)/1e9
                        data.append([current_time, int(temperature*60)])
                except (IndexError, json.JSONDecodeError, TypeError, ValueError):
                    continue  # skip malformed lines
    df = pd.DataFrame(data, columns=['Time (s)', 'Power (W)'])
    return df


import sqlite3
import pandas as pd

def write_layers(db_path, layer_data_df):
    # Your DataFrame: layer_data_df  (columns: "Time (s)", "Layer", "Build height (mm)")

    conn = sqlite3.connect(db_path)
    layer_data_df.to_sql("layer_metrics", conn, if_exists="replace", index=False)

    conn.close()

def write_energy_input(db_path, energy_data_df):
    # Your DataFrame: energy_data_df  (columns: "Time (s)", "Power (W)")

    conn = sqlite3.connect(db_path)
    energy_data_df.to_sql("energy_input", conn, if_exists="replace", index=False)

    conn.close()

def write_temperatures(db_path, temp_data_df):
    # Your DataFrame: energy_data_df  (columns: "Time (s)", "Sensor1", "Sensor2", ...)

    conn = sqlite3.connect(db_path)
    temp_data_df.to_sql("temperatures", conn, if_exists="replace", index=False)

    conn.close()


import matplotlib.pyplot as plt

def plot_obp_img(img):
    try:
        plt.imshow(img, origin='upper', cmap='gray')
        plt.title("Rasterized coverage")
        plt.axis('equal'); plt.show()
    except Exception:
        pass


import pandas as pd
import json

def get_recoate_times(log_file):
    # List to collect the extracted data
    data = []
    with open(log_file, 'r') as f:
        for line in f:
            if 'freemelt/0/BuildProcess/0/BuildStatus/Trace/Activity' in line:
                try:
                    json_part = line.split(' ', 1)[1]
                    values = json.loads(json_part)

                    temperature = values.get('current_activity')
                    if temperature.startswith("Recoat cycle"):
                        parts = temperature.split("Layer")
                        if len(parts) > 1:
                            layer_number = parts[1].strip()
                            time_ns = int(values.get('time'))  # Make sure it's an int
                            data.append([time_ns, int(layer_number)-1])
                except (IndexError, json.JSONDecodeError, TypeError, ValueError):
                    continue  # skip malformed lines


    # Convert to DataFrame
    df = pd.DataFrame(data, columns=['Time (s)', 'Layer'])

    return df


from dataclasses import dataclass, asdict, field
from typing import List, Any, Mapping
import json


@dataclass
class FileStats:
    nmb_repetitions: int #n
    time: float #s
    energy: float #J
    nmb_spots: int #n
    line_length: float #m

@dataclass
class LayerInfo:
    layer_index: int
    layer_height: float #mm
    recoate_time: float #s
    files: List[FileStats]

    def get_layer_time(self):
        return sum(file.time * file.nmb_repetitions for file in self.files) + self.recoate_time
    def get_layer_energy(self):
        return sum(file.energy * file.nmb_repetitions for file in self.files)
    
@dataclass
class BuildInfo:
    layers: List[LayerInfo]
    start_temp: float #degress C
    start_heat: List[FileStats]

    def get_total_duration(self):
        total = 0.0
        for layer in self.layers:
            total += layer.get_layer_time()
        return total

    def to_json(self, file_path: str):
        """Serialize the BuildInfo object to a JSON file."""
        with open(file_path, "w") as f:
            json.dump(asdict(self), f, indent=4)

    @staticmethod
    def from_json(file_path: str) -> 'BuildInfo':
        """Deserialize a JSON file into a BuildInfo object."""
        with open(file_path, "r") as f:
            data = json.load(f)
        
        layers = [
            LayerInfo(
                layer_index=l['layer_index'],
                layer_height=l['layer_height'],
                recoate_time=l['recoate_time'],
                files=[FileStats(**fs) for fs in l['files']]
            )
            for l in data['layers']
        ]
        start_heat = [FileStats(**fs) for fs in data['start_heat']]
        return BuildInfo(
            layers=layers,
            start_temp=data['start_temp'],
            start_heat=start_heat
        )
@dataclass
class GeometryFileInfo:
    melt_area_mm2: float  # mm2
    total_area_mm2: float # mm2
    spot_size_um: float   # um

    @staticmethod
    def from_dict(d: Mapping[str, Any]) -> "GeometryFileInfo":
        # Cast to float in case the JSON has ints/strings
        return GeometryFileInfo(
            melt_area_mm2=float(d["melt_area_mm2"]),
            total_area_mm2=float(d["total_area_mm2"]),
            spot_size_um=float(d["spot_size_um"]),
        )

@dataclass
class GeometryLayerInfo:
    layer_index: int
    melt_area_mm2: float  # mm2
    melt_portion: float   # %
    # Make files optional in JSON; write as [] when missing
    files: List[GeometryFileInfo] = field(default_factory=list)

    @staticmethod
    def from_dict(d: Mapping[str, Any]) -> "GeometryLayerInfo":
        files_raw = d.get("files") or []  # allow missing/null -> []
        files_parsed = [
            GeometryFileInfo.from_dict(f) if not isinstance(f, GeometryFileInfo) else f
            for f in files_raw
        ]
        return GeometryLayerInfo(
            layer_index=int(d["layer_index"]),
            melt_area_mm2=float(d["melt_area_mm2"]),
            melt_portion=float(d["melt_portion"]),
            files=files_parsed,
        )

@dataclass
class GeometryInfo:
    layers: List[GeometryLayerInfo]

    def to_json_file(self, filepath: str, indent: int = 4) -> None:
        """Write the GeometryInfo object to a JSON file."""
        with open(filepath, "w", encoding="utf-8") as f:
            json.dump(asdict(self), f, indent=indent)

    @classmethod
    def from_json_file(cls, filepath: str) -> "GeometryInfo":
        """Read a GeometryInfo object from a JSON file.

        Accepts either:
          - {"layers": [ ...layer objects... ]}
          - [ ...layer objects... ]
        Each layer object may include "files": [ ...file objects... ] or omit it.
        If omitted or null, 'files' becomes an empty list.
        """
        with open(filepath, "r", encoding="utf-8") as f:
            data = json.load(f)

        if isinstance(data, dict):
            layers_data = data.get("layers", []) or []
        elif isinstance(data, list):
            layers_data = data
        else:
            raise ValueError("Invalid JSON format: expected object with 'layers' or a list.")

        layers = [GeometryLayerInfo.from_dict(item) for item in layers_data]
        return cls(layers=layers)


import unittest

import obanalyser.analyse_obf_geometry as analyse_obf_geometry
import obanalyser.data_classes as data_classes

class TestMain(unittest.TestCase):
    def test_main(self):
        path1 = r"tests\input\cubes_test\buildInfo.json"
        path2 = r"tests\output\geometryInfo.json"

        geometry_info = analyse_obf_geometry.analyse_obf_geometry(path1)
        geometry_info.to_json_file(path2)
        print(f"Geometry analysation saved to {path2}")

        geometry_info2 = data_classes.GeometryInfo.from_json_file(path2)
        


if __name__ == "__main__":
    unittest.main()






import pandas as pd
import json

def get_layer_heights(log_file, t0=0):
    # List to collect the extracted data
    data = []
    z0 = None
    with open(log_file, 'r') as f:
        for line in f:
            if 'freemelt/0/RecoaterService/0/Recoater/Position/BuildServo' in line:
                try:
                    json_part = line.split(' ', 1)[1]
                    values = json.loads(json_part)

                    height = round(values.get('Location'), 3)
                    if z0 is None:
                        z0 = height
                        height = 0
                    else:
                        height = z0-height
                    time = int(values.get('time'))
                    if time >= t0:
                        data.append([height, (time-t0)/1e9])
                except (IndexError, json.JSONDecodeError, TypeError, ValueError):
                    continue  # skip malformed lines


    # Convert to DataFrame
    df = pd.DataFrame(data, columns=['Build height (mm)', 'time'])
    df_unique = df.drop_duplicates(subset="Build height (mm)", keep="first").reset_index(drop=True)

    return df_unique


import pandas as pd
import numpy as np
from functools import reduce

from obanalyser.obr.recoate_times import get_recoate_times
from obanalyser.obr.layer_heights import get_layer_heights
from obanalyser.obr.write_db import write_layers, write_energy_input, write_temperatures
from obanalyser.obr.exposure import get_power_times
from obanalyser.obr.temperatures import read_mqqt

def analyse_obr(
        log_file, 
        db_path=r"tests\output\obr_analysis.db",
        temp_sensors=[
            ('freemelt/0/ChamberService/0/BuildTemperature/Name/Sensor1', 'Temperature'),
            ('freemelt/0/PyrometerService/0/BuildTemperature/Name/Pyrometer',"Temperature")
        ]):
    # Get recoate times
    recoate_df = get_recoate_times(log_file)
    t0 = recoate_df["Time (s)"].iloc[0]
    recoate_df["Time (s)"] = (recoate_df["Time (s)"] - t0) / 1e9  # Convert to seconds
    # Get build heights
    heights = get_layer_heights(log_file, t0=t0)

    df_build = heights # with columns ["Build height (mm)", "time"]
    df_layers = recoate_df #with columns ["Time (s)", "Layer"]

    bins = np.append(df_layers["Time (s)"].values, np.inf)  # add infinity for the last layer
    df_build["Layer"] = pd.cut(
        df_build["time"],
        bins=bins,
        labels=df_layers["Layer"],
        right=False  # intervals are [start, end)
    )
    df_result = df_build[["Build height (mm)", "Layer"]].dropna().reset_index(drop=True)
    df_last = df_result.groupby("Layer", as_index=False).last()
    layer_combined = pd.merge(recoate_df, df_last, on="Layer")
    layer_combined["Layer height (mm)"] = layer_combined["Build height (mm)"].diff().fillna(layer_combined["Build height (mm)"].iloc[0])
    write_layers(db_path, layer_combined)
    # Get exposure times
    exposure_times = get_power_times(log_file, t0)
    write_energy_input(db_path, exposure_times)
    # Temperatures
    dfs = []
    for sensor_tag, sensor_name in temp_sensors:
        temp = read_mqqt(log_file, sensor_tag, t0, value_key=sensor_name)
        dfs.append(temp)
    dfs_long = []  # to collect transformed DFs
    for i, df in enumerate(dfs):
        # Extract sensor name (the second column)
        sensor_col = df.columns[1]
        # Rename columns to consistent names
        temp_df = df.rename(columns={
            "Time (s)": "Time",
            sensor_col: "Temp"
        }).copy()
        # Add Index column (sensor index or name)
        temp_df["Index"] = sensor_col  # or i if you prefer numbers
        dfs_long.append(temp_df)
    df_temps = pd.concat(dfs_long, ignore_index=True)
    write_temperatures(db_path, df_temps)



import json
import pandas as pd

def read_mqqt(log_file, tag, t0, value_key='Temperature', time_key='time'):
    # File paths

    # List to collect the extracted data
    data = []
    # Read and extract temperature/time
    with open(log_file, 'r') as f:
        for line in f:
            if line.split(" ", 1)[0] == tag:
                try:
                    json_part = line.split(' ', 1)[1]
                    values = json.loads(json_part)
                    #print(values)
                    temperature = values.get(value_key)
                    time = int(values.get(time_key))  # Make sure it's an int
                    if time > t0:
                        current_time = (time - t0)/1e9
                        data.append([current_time, temperature])
                except (IndexError, json.JSONDecodeError, TypeError, ValueError):
                    continue  # skip malformed lines

    # Normalize time: convert nanoseconds to seconds and set first to 0
    last_part = tag.split("/")[-1]
    df = pd.DataFrame(data, columns=['Time (s)', last_part])
    return df